static char sccsid[] = "@(#) dn2app.cc 1.4 98/09/22 13:56:58";

// The next two lines are only to ensure bool gets defined appropriately.
#define SP_MULTI_BYTE

#include <iostream>
#include <strstream>
#include "config.h"
#include "Boolean.h"
#include "ParserEventGeneratorKit.h"

#include "OutputCharStream.h"

#include "dn2app.h"

ostrstream& operator<< (ostrstream &os, SGMLApplication::CharString s)
{
    const Char *p = s.ptr;
    size_t n = s.len;
    for (; n > 0; p++, n--) {
	switch (*p) {
	  case '\\':			// FIXME we're punning Chars and chars
	    os << "\\\\";
	    break;
	  case '\n':
	    os << '\n';
	    break;
	  case '\r':
	    os << "\\n";
	    // if (outputLine_ && haveData_)
	    // lastLineno_++;
	    break;
	  default:
	    // FIXME not clear what to do here given possibility of wide characters
	    if (*p < 040) {
		static const char digits[] = "0123456789";
		os << "\\0" << digits[*p / 8] << digits[*p % 8];
	    }
	    else
	      os.put((unsigned char)*p);
	    break;
	}
    }
    return os;
}

ostream& operator<< (ostream &os, SGMLApplication::CharString s)
{
    const Char *p = s.ptr;
    size_t n = s.len;
    for (; n > 0; p++, n--) {
	switch (*p) {
	  case '\\':			// FIXME we're punning Chars and chars
	    os << "\\\\";
	    break;
	  case '\n':
	    os << "\n";
	    break;
	  case '\r':
	    os << "\\n";
	    // if (outputLine_ && haveData_)
	    // lastLineno_++;
	    break;
	  default:
	    // FIXME not clear what to do here given possibility of wide characters
	    if (*p < 040) {
		static const char digits[] = "0123456789";
		os << "\\0" << digits[*p / 8] << digits[*p % 8];
	    }
	    else
	      os.put((unsigned char)*p);
	    break;
	}
    }
    return os;
}

ostream &operator<<(ostream &os, SGMLApplication::CharString *s)
{
    cerr << "in << with CharString *" << endl;
    const Char *p = s->ptr;
    size_t n = s->len;
    for (; n > 0; p++, n--) {
	switch (*p) {
	  case '\\':			// FIXME we're punning Chars and chars
	    os << "\\\\";
	    break;
	  case '\n':
	    os << "\n";
	    break;
	  case '\r':
	    os << "\\n";
	    // if (outputLine_ && haveData_)
	    // lastLineno_++;
	    break;
	  default:
	    // FIXME not clear what to do here given possibility of wide characters
	    if (*p < 040) {
		static const char digits[] = "0123456789";
		os << "\\0" << digits[*p / 8] << digits[*p % 8];
	    }
	    else
	      os.put((unsigned char)*p);
	    break;
	}
    }
    return os;
}


DN2Application::DN2Application() : depth_(0) {
    theDoc = new DocumentData;
}

DN2Application::~DN2Application() {
    delete theDoc;
}

void DN2Application::startElement(const StartElementEvent &event) {
    // Add node to DocumentData
    theDoc->start_node(event.gi);
    for (unsigned i = 0; i< event.nAttributes; i++) {
	// cout << event.attributes[i].name;
	switch (event.attributes[i].type)
	  {
	    case Attribute::invalid:
	    case Attribute::implied:
	      break;
	    case Attribute::cdata:
	      cout << " Attribute contains cdata" << endl;
	      cout << " #" << event.attributes[i].nCdataChunks;
	      for (unsigned j=0; j<event.attributes[j].nCdataChunks; j++)
		cout << ": " << event.attributes[j].cdataChunks[j].data;
	      break;
	    case Attribute::tokenized:
	      theDoc->add_attribute(&(event.attributes[i].name),&(event.attributes[i].tokens));
	      break;
	    default:
	      cout << "Shouldn't reach here!!!" << endl;
	      break;
	  }
    }
    depth_++;
}


void DN2Application::startDtd( const StartDtdEvent &) {
  cout << "Started DTD" << endl;
}

void DN2Application::endDtd(const EndDtdEvent &) {
  cout << "Ended Dtd" << endl;
}

void DN2Application::data(const DataEvent &event) {
  theDoc->add_data(&(event.data));
}

void DN2Application::sdata(const SdataEvent &) {
  cout << "SData Event" << endl;
}
  
void DN2Application::endElement(const EndElementEvent &event)
{
  depth_--;
  theDoc->end_node();
}

void DN2Application::error(const ErrorEvent &event) {

    switch(event.type) {
      case ErrorEvent::info:
	cout << "!!INFO!!" << endl;
	break;
      case ErrorEvent::warning:
	cout << "!!Warning!!" << endl;
	break;
      case ErrorEvent::quantity:
      case ErrorEvent::idref:
      case ErrorEvent::capacity:
      case ErrorEvent::otherError:
	throw SGMLError(event);
	break;
      default:
	cout << "Ack! Why did we get here?" << endl;
	break;
    }
}

void DN2Application::cleanup() {
    // Received SGMLerror now trying to clean up
    theDoc->error_cleanup();
}

void DN2Application::PrintTree() {
    cout << theDoc->return_dn2();
}

string DN2Application::GetDN2() {
    return theDoc->return_dn2();
}

string DN2Application::GetTclDN2() {
    return theDoc->return_tcldn2();
}

SGMLError::SGMLError(const SGMLApplication::ErrorEvent &e) {
    // Extract the line number and column from the error event 
    // these will be passed back to the calling routine.
    ostrstream os;
    os << e.message;
    cerr << "e.message =" << e.message << endl;
    errorstring = os.str();
    // print out the errormessage
    cerr << "errorstring =" <<  errorstring << endl;
}

