#ifdef RCSID
static char rcsid[] = "$Header: /home/smart/release/src/smart/smart.c,v 11.2 1993/02/03 16:48:11 smart Exp $";
#else
static char sccsid[] = "@(#)dn2smart.c	1.2";
#endif

/* Copyright (c) 1991, 1990, 1984 - Gerard Salton, Chris Buckley. 

   Permission is granted for use of this file in unmodified form for
   research purposes. Please contact the SMART project to obtain 
   permission for other uses.
*/

/* provide MyRetrieveDocuments() and MyPresentDoc() to the Logicon code.
 * These routines interface to a smart that translates the presented dn2tosend
 * into willow internal form, and then hands off to the willow z39.50 client.
 *
 * willow uses recinfo_output() to do the writing to file out
 */

#define DEBUG
#define DEBUG2

#include <ctype.h>
#include "common.h"
#include "param.h"
#include "io.h"
#include "functions.h"
#include "smart_error.h"
#include "proc.h"
#include "spec.h"
#include "docindex.h"
#include "trace.h"
#include "accounting.h"
#include "buf.h"
#include "infoneed.h"
#include "infoneed_parse.h"

#include "dn2smart.h"

#define COLL_LIST_NAME  "./coll_list"

static char specname[] = "spec.logicon";
static int initialized_flag = -1;

static COLL_LIST_ENTRY *coll_list;
static int num_coll_list = 0;
static int current_coll_index = -1;

static INT_RETRIEVED_DOC *int_ret_doc;
static int num_int_ret_doc = -1;

static RETRIEVED_DOC *ret_doc;
static int num_ret_doc = -1;
static int max_num_ret_doc = -1;

static SM_BUF doc_buf;

static char tmp_buf[PATH_LEN];
static char new_dn2[PATH_LEN];

#define put_one_error_message(dest,format,message) {(void) sprintf (tmp_buf, format, message); *dest = tmp_buf;}

RETRIEVED_DOC *
MyRetrieveDocuments(coll_name, dn2tosend, status, error_msg, dn2returned,
                                                                num_retrieved)
char *coll_name;
char *dn2tosend;
int *status;
char **error_msg;
char **dn2returned;
int *num_retrieved;
{
    char *error_message;
    int coll_index;
    int i;

    /* If first call to RetrieveDocuments, initialize everything */
    if (-1 == initialized_flag) {

        /* Get list of valid coll_names */
        if (-1 == get_coll_list (COLL_LIST_NAME, &coll_list, &num_coll_list)) {
            put_one_error_message(error_msg,
                                  "Unable to get coll_list %s.\n",
                                  COLL_LIST_NAME);
            *status = 1;
            return (NULL);
        }
#ifdef DEBUG
        fprintf (stderr, "Initialized coll list");
#endif
        if (-1 == init_smart (specname)) {
            put_one_error_message(error_msg,
                                  "Unable to initialize smart, spec %s.\n",
                                  specname);
            *status = 1;
            return (NULL);
        }
#ifdef DEBUG
        fprintf (stderr, "...Initialized smart interface\n");
#endif
        doc_buf.size = 0;
        initialized_flag = 1;
    }

    for (coll_index = 0;
         coll_index < num_coll_list &&
             0 != strcmp (coll_list[coll_index].coll_name, coll_name);
         coll_index++) 
        ;
    if (coll_index >= num_coll_list) {
        put_one_error_message(error_msg,
                              "Illegal collection %s.\n",
                              coll_name);
        *status = 2;
#ifdef DEBUG
        fprintf (stderr, "Missing collection name %s - returned error\n", coll_name);
#endif
        return (NULL);
    }
#ifdef DEBUG
        fprintf (stderr, "Found collection name %s\n", coll_name);
#endif

    if (coll_index != current_coll_index) {
        /* Must open new collection after possibly closing old one */
        if (current_coll_index != -1) {
            /* No code needed currently */
        }
        switch (coll_list[coll_index].coll_type) {
        case CLTYPE_SMART:
              /* Ignore for now */
              break;
        case CLTYPE_Z3950:
              if (-1 == open_z3950 (&coll_list[coll_index])) {
                  put_one_error_message(error_msg,
                                    "Collection not currently available %s.\n",
                                    coll_name);
                  *status = 3;
                  return (NULL);
              }
              break;
        default:
              put_one_error_message(error_msg,
                                "Illegal collection type %s.\n",
                                coll_name);
              *status = 3;
              return (NULL);
              break;
        }
        current_coll_index = coll_index;
#ifdef DEBUG
        fprintf (stderr, "Opened collection \n");
#endif
    }

    /* Run query */
    switch (coll_list[coll_index].coll_type) {
    case CLTYPE_SMART:
        /* Ignore for now */
#ifdef notdef
        if (-1 == run_smart_query (dn2tosend,
                                   coll_name,
                                   &int_ret_docs,
                                   &num_int_ret_docs,
                                   dn2returned,
                                   &error_message)) {
            put_one_error_message(error_msg,
                                  "%s\n",
                                  error_message);
            *status = 4;
            return (NULL);
        }
#endif  
        break;
    case CLTYPE_Z3950:
        if (-1 == run_z3950_query (dn2tosend,
                                   coll_name,
                                   &int_ret_doc,
                                   &num_int_ret_doc,
                                   &error_message)) {
            put_one_error_message(error_msg,
                                  "%s\n",
                                  error_message);
            *status = 4;
#ifdef DEBUG
            fprintf (stderr, "Return error running Z3950 query\n");
#endif
            return (NULL);
        }
        (void) strncpy (new_dn2, dn2tosend, PATH_LEN);
        *dn2returned = new_dn2;
        break;
    default:
        put_one_error_message(error_msg,
                              "Illegal collection type %s.\n",
                              coll_name);
        *status = 1;
        return (NULL);
        break;
    }
#ifdef DEBUG
        fprintf (stderr, "Ran query \n");
#endif

    /* Copy internal form of docs to list to be returned */
    /* Note: separate internal form now since eventually will have different */
    /* internal lists from multiple collections to be merged. */
    if (num_int_ret_doc > max_num_ret_doc) {
        if (max_num_ret_doc > 0)
            (void) free ((char *) ret_doc);
        if (NULL == (ret_doc = (RETRIEVED_DOC *) malloc ((unsigned) num_int_ret_doc * sizeof (RETRIEVED_DOC)))) {
            put_one_error_message(error_msg,
                                  "Out of space %s.\n",
                                  "RetrieveDocuments");
            *status = 1;
            return (NULL);
        }
        max_num_ret_doc = num_int_ret_doc;
    }
    for (i = 0; i < num_int_ret_doc; i++) {
#ifdef DEBUG2
        fprintf (stderr, "    Retrieved Doc %s, %s, %f\n",
                 int_ret_doc[i].external_name,
                 int_ret_doc[i].title,
                 int_ret_doc[i].score);
#endif
        ret_doc[i].external_id = int_ret_doc[i].external_name;
        ret_doc[i].title = int_ret_doc[i].title;
        ret_doc[i].score = int_ret_doc[i].score;
        ret_doc[i].hilight_info = int_ret_doc[i].hilight_info;
    }
    num_ret_doc = num_int_ret_doc;

    /* Return with success */
    *status = 0;

    *num_retrieved = num_ret_doc;
#ifdef DEBUG
        fprintf (stderr, "Query Success: %4d docs returned\n", num_ret_doc);
#endif

    return (&ret_doc[0]);
}

char *
MyPresentDoc(external_id, status, error_msg)
char *external_id;
int *status;
char **error_msg;
{
    int i;
    char *error_message;

    if (-1 == initialized_flag)         /* not running smart */
        return (char *)0;
    
#ifdef DEBUG
        fprintf (stderr, "PresentDoc called for %s\n", external_id);
#endif

    /* Find document in current list of retrieved docs */
    for (i = 0; i < num_int_ret_doc; i++) {
        if (0 == strcmp (external_id, int_ret_doc[i].external_name))
            break;
    }
    if (i >= num_int_ret_doc) {
        put_one_error_message(error_msg,
                              "Document not in current list %s.\n",
                              external_id);
        *status = 3;
#ifdef DEBUG
        fprintf (stderr, "Doc %s not found: error return\n", external_id);
#endif
        return (NULL);
    }
#ifdef DEBUG
    fprintf (stderr, "Internal form of Doc %s found\n", external_id);
#endif
        
    doc_buf.end = 0;

    switch (int_ret_doc[i].coll_type) {
    case CLTYPE_SMART:
        /* Ignore for now */
#ifdef notdef
        if (-1 == show_smart_doc (int_ret_doc[i].internal_name,
                                  &doc_buf,
                                  &error_message)) {
            put_one_error_message(error_msg,
                                  "%s\n",
                                  error_message);
            *status = 4;
            return (NULL);
        }
#endif  
        break;
    case CLTYPE_Z3950:
        if (-1 == show_z3950_doc (i+1, &doc_buf, &error_message)) {
            put_one_error_message(error_msg,
                                  "%s\n",
                                  error_message);
            *status = 4;
#ifdef DEBUG
            fprintf (stderr, "Doc %s not found: Z3950 error return %s\n", external_id, error_message);
#endif
            return (NULL);
        }
        break;
    default:
        put_one_error_message(error_msg,
                              "Illegal collection type %s for doc.\n",
                              "");
        *status = 1;
        return (NULL);
        break;
    }
#ifdef DEBUG
            fprintf (stderr, "Doc %s Returned\n", external_id);
#endif
#ifdef DEBUG2
            fwrite (doc_buf.buf, 1, doc_buf.end, stderr);
#endif

    return (doc_buf.buf);
}
