static char sccsid[] = "@(#)smart_willow.c	1.1";
/*
 * These routines interface to a smart that translates the presented dn2tosend
 * into willow internal form, and then hands off to the willow z39.50 client.
 *
 * willow uses recinfo_output() to do the writing to file out
 */

#include <ctype.h>
#include "common.h"
#include "param.h"
#include "io.h"
#include "functions.h"
#include "smart_error.h"
#include "proc.h"
#include "spec.h"
#include "docindex.h"
#include "trace.h"
#include "accounting.h"
#include "infoneed.h"
#include "infoneed_parse.h"

#include "dn2smart.h"

int init_smart(), close_smart();
int smart_dn2_convert();

static int dn2_inst = -1;

#include <stdarg.h>
#include <sys/stat.h>
#include "z39constant.h"
#include "api.h"
#include "msg.h"
#include "utility.h"

static void close_command();

void init_ifn();
int infoneed_add_child();
void print_winfoneed();
int init_winfoneed_parse(), winfoneed_parse(), close_winfoneed_parse();

static char *usage = "Usage: smart action spec [spec_param spec_value]*";

static char *trace_file;
static char *log_file_name;
static char *msg_file;

/* for willow code */
int debug;
FILE *in;
FILE *out;
FILE *log_file;
FILE *msg_log;
SESSION *session;

/* next section from willow client.c */

#define MAX_BUFFER          2048           /* Maximum character buffer size */

/* Used by the assignment subroutine below */

#define NUMBER              1              /* Integer parameters */
#define P_WORD              2              /* A word parameter */
#define P_WORD_RAW          3              /* A raw word (without quote processing) */
#define P_PHRASE            4              /* A phrase parameter */


static SPEC_PARAM spec_args[] = {
    PARAM_LONG( "global_trace_start", &global_trace_start ),
    PARAM_LONG( "global_trace_end",   &global_trace_end ),
    PARAM_LONG( "global_start",       &global_start ),
    PARAM_LONG( "global_end",         &global_end ),
    PARAM_LONG( "global_accounting",  &global_accounting ),
    PARAM_DBFILE( "global_trace_file",&trace_file ),
    PARAM_LONG( "willow.debug",       &debug ),
    PARAM_DBFILE( "willow.logfile",   &log_file_name ),
    PARAM_DBFILE( "willow.msgfile",   &msg_file ),
    TRACE_PARAM ("smart.trace")
    };
static int num_spec_args = sizeof (spec_args) / sizeof (spec_args[0]);

/* Declaration of the root of the procedure hierarchy. Here so the
   entire hierarchy will be loaded from libraries. */
extern TAB_PROC_TAB proc_root[];
extern int num_proc_root;
TAB_PROC_TAB proc_root_root = {
    "root",     TPT_TAB,        proc_root,      NULL,   &num_proc_root
    };
static SPEC spec;

int init_smart (spec_file)
char *spec_file;
{
    /*SPEC spec;*/

    /* Read the specification file given by the second argument */
    if (UNDEF == get_spec (spec_file, &spec)) {
        print_error ("smart: Illegal spec file", "Quit");
        exit (2);
    }

    /* Lookup the values of the relevant parameters */
    if (UNDEF == lookup_spec (&spec, &spec_args[0], num_spec_args)) {
        print_error ("smart: Global parameters not found", "Quit");
        exit (3);
    }

    /* allow tracing */
    SET_TRACE (0);
#ifdef TRACE
    if (VALID_FILE (trace_file)) {
        if (NULL == (global_trace_fd = fopen (trace_file, "w"))) {
            print_error ("smart: Trace file not accessible", "Quit");
            exit (3);
        }
    }
#endif
    
    PRINT_TRACE (2, print_string, "Trace: entering smart");
    PRINT_TRACE (4, print_spec, &spec);

    /* do initialization for willow */
    if (VALID_FILE (log_file_name)) {
        if (NULL == (log_file = fopen (log_file_name, "a"))) {
            print_error ("smart: Willow log file not accessible", "Quit");
            exit (3);
        }
    } else
        log_file = 0;
    if (VALID_FILE (msg_file)) {
        if (NULL == (msg_log = fopen (msg_file, "a"))) {
            print_error ("smart: Willow message file not accessible", "Quit");
            exit (3);
        }
    } else
        msg_log = 0;
    in = stdin;
    out = stdout;
    session = NULL;
    
    /* initialize the dn2 conversion procedure */
    if (UNDEF == (dn2_inst = init_infoneed_parse (&spec, NULL))) {
        print_error ("smart Initialization:", "Quit");
        return (-1);
    }

    return (1);
}


int
close_smart()
{
    /* close willow */
    close_command(&session, (char *)0);
    if (VALID_FILE (log_file_name))
        (void) fclose (log_file);
    if (VALID_FILE (msg_file))
        (void) fclose (msg_log);

    if (UNDEF == close_infoneed_parse (dn2_inst)) {
        print_error ("smart: Close:", "Quit");
        return (-1);
    }

    PRINT_TRACE (2, print_string, "Trace: leaving smart");
#ifdef TRACE
    if (VALID_FILE (trace_file))
        (void) fclose (global_trace_fd);
#endif
    return (0);
}

int
smart_dn2_convert (dn2_in, new_query)
char *dn2_in;
char new_query[];
{
    IFN_TREE_NODE *ifn_prev_child, *ifn_this_child;
    IFN_TREE_NODE ifn_root;
    IFN_TREE ifn_tree;
    IFN_QUERY ifn_query;
    SM_BUF outbuf;

    outbuf.size = outbuf.end = 0;

    init_ifn (&ifn_root);
    ifn_query.query_text = dn2_in;

    ifn_root.op = IFN_ROOT_OP;
    ifn_prev_child = ifn_this_child = (IFN_TREE_NODE *)0;

    /* Construct a query with dn2_in as text, and parse it */
    if (!strncmp(dn2_in, "<!", 2)) {
        char *cp = strchr(dn2_in, '>');
        ifn_query.start_pos = cp - dn2_in + 1;
    } else
        ifn_query.start_pos = 0;
    ifn_query.end_pos = strlen(dn2_in);
    if (ifn_query.end_pos > ifn_root.text_end)
        ifn_root.text_end = ifn_query.end_pos;
    ifn_query.pos = ifn_query.start_pos;
    ifn_query.ifn_inst = dn2_inst;

    if (UNDEF == infoneed_add_child(&ifn_root, &ifn_prev_child,
                                &ifn_this_child, ifn_query.pos))
        return(UNDEF);

    if (UNDEF == infoneed_parse (&ifn_query, ifn_this_child, dn2_inst))
        return(UNDEF);

    ifn_tree.ifn_node = &ifn_root;
    ifn_tree.query_text = dn2_in;
    print_winfoneed (&ifn_tree, &outbuf);
    if (outbuf.end > MAX_BUFFER)
        outbuf.end = MAX_BUFFER;
    strncpy (new_query, outbuf.buf, outbuf.end);
    new_query[outbuf.end] = '\0';

    PRINT_TRACE (4, print_string, new_query);

    return (1);
}


/* next section from willow client.c */

/*--------------------------------------------------------------*
 |                    CLIENT_MSG                                |
 | Function:                                                    |
 |    Print out the message.                                    |
 | Parameters:                                                  |
 |    session  : Pointer to session information.                |
 |    format   : Format for the message.                        |
 |    va_list  : Arguements for the message.                    |
 | Returns:                                                     |
 |    Nothing.                                                  |
 | Description:                                                 |
 |    Uses vfprintf to print out the message.  Also checks the  |
 |    messages saved in IRI and prints them out.                |
 | Note:                                                        |
 |    Should probably just redirect io to stdin and stdout.     |
 *--------------------------------------------------------------*/

static void client_msg(char *format,
                       ...)
{
    extern FILE *msg_log;
    va_list more;

    va_start(more, format);
    vfprintf(out, format, more);
    if (msg_log)
        {
            vfprintf(msg_log, format, more);
            fflush(msg_log);
        }
    msg_output();
    fflush(out);

    va_end(more);
}

/*--------------------------------------------------------------*
 |                    ASSIGNMENT                                |
 | Function:                                                    |
 |    Assign a the next word in the string.                     |
 | Parameters:                                                  |
 |    input   : Address of input string.                        |
 |    name    : Pointer to assignment name.                     |
 |    type    : Type of value to assign.                        |
 |    assign  : Pointer to string to hold assignment            |
 | Returns:                                                     |
 |    length of assigned string.                                |
 | Description:                                                 |
 |                                                              |
 | Note:                                                        |
 |                                                              |
 *--------------------------------------------------------------*/

static int assignment(char **input,
                      char *name,
                      int  type,
                      char *assign)
{
    char compare[MAX_BUFFER];
    char *token;
    int  length;
    char *phrase = *input;

    /* Check if a name needs to be matched */

    if (*name != EOS)
        {
            length = tokenize(&phrase, &token, " =");

            UNQUOTE(token, length);

            uppercase(compare, token, length);
            if (strncmp(name, compare, length) != 0)
                {
                    /* No match so set to no value */
 
                    if (assign != NULL)
                        {
                            if (type == NUMBER)
                                *((int *) assign) = 0;
                            else
                                assign[0] = EOS;
                        }
                    return 0;
                }
        }

    if (assign != NULL)
        {
            length = tokenize(&phrase, &token, " =");

            if (*name != EOS && *token == EQUAL_SIGN)
                {
                    phrase++;
                    length = tokenize(&phrase, &token, " ");
                }
            
            switch (type)
                {
                    case NUMBER:
                        if (length <= 0)
                           *((int *) assign) = 0;
                        else
                           *((int *) assign) = atoi(token);
                        break;

                    case P_WORD:

                        UNQUOTE(token, length);

                        /* Fall through into P_WORD_RAW */

                    case P_WORD_RAW:

                        if (length <= 0)
                            assign[0] = EOS;
                        else
                            {
                                COPY(assign, token, length);
                                assign[length] = EOS;
                            }
                        break;

                    case P_PHRASE:
                        length = strlen(token);
                        phrase = token + length;   /* Point to EOS */

                        /* Trim trailing blanks */

                        while (token[length - 1] == BLANK && length > 1) length--;
                        
                        COPY(assign, token, length);
                        assign[length] = EOS;
                        break;

                    default:
                        break;
                }
        }

    *input = phrase;
    return length;
}

/* return -1 on error */
/* revised willow select_command() */
int
open_z3950(coll)
COLL_LIST_ENTRY *coll;
{
    char port[MAX_BUFFER];
    char host[MAX_BUFFER];
    char globalset[MAX_BUFFER];
    char id[MAX_BUFFER];
    char syntax[MAX_BUFFER];
    int  code, retcode;
    char *name, *parameters;

    name = coll->coll_name;
    parameters = coll->parameters;

    assignment(&parameters, "HOST", P_WORD, host);
    assignment(&parameters, "PORT", P_WORD, port);
    assignment(&parameters, "GLOBALSET", P_WORD, globalset);
    assignment(&parameters, "ID", P_WORD, id);
    assignment(&parameters, "SYNTAX", P_WORD, syntax);

#ifdef MARKER 
    assignment(&parameters, "MARC", NUMBER, &marc_type);
#endif

    while (*parameters && *parameters == BLANK) parameters++;

    if (name[0] == EOS) {
        client_msg("-ERROR: Missing database name in SELECT command\n");
        retcode = -1;
    } else if (*parameters != EOS) {
        client_msg("-ERROR: Extra paramaters on SELECT: %s\n", parameters);
        retcode = -1;
    } else {
        code = z39_initial(&session, name, host, port, globalset, syntax, id);

        if (code) {
            client_msg("-ERROR: Unable to establish connection.\n");
            retcode = -1;
        } else {
#ifdef MARKER
            (*p_session)->marc_type = marc_type;
#endif 
            client_msg("-OK: Connection to %s is established.\n", 
                           session_host(session));
            retcode = 0;
        }
    }
    return(retcode);
}

/*--------------------------------------------------------------*
 |                    CLOSE_COMMAND                             |
 | Function:                                                    |
 |    Close the connection to the Z39.50 server.                |
 | Parameters:                                                  |
 |    session    : Pointer to session information.              |
 |    parameters : Character string for select command.         |
 | Returns:                                                     |
 |    0 is successful, -1 if failed.                            |
 | Description:                                                 |
 |    Interface to call ir_close to close the Z39.50 server     |
 |    connection.                                               |
 | Note:                                                        |
 |                                                              |
 *--------------------------------------------------------------*/

static void close_command(SESSION **p_session,
                          char *parameters)
{
    int code;

    code = z39_close(p_session, parameters);

    if (code)
        client_msg("-ERROR: On closing connection\n");
    else
        client_msg("-OK: Connection was closed\n");
}

#define WILLOW_TEMP "/tmp/willow_doc"

/* some resemblance to willow client.c display_command() */
int
show_z3950_doc(pos, docbuf, error_message)
int pos;
SM_BUF *docbuf;
char **error_message;
{
    RECORD_INFO *info;
    int  code, retcode;
    FILE *tmp_file;
    struct stat stat_buf;
    char *text;

    /* the willow library will write the document to file 'out' */
    tmp_file = fopen(WILLOW_TEMP, "w");
    if ((FILE *)0 == tmp_file) {
        *error_message = "couldn't open willow temp file";
        return(-1);
    }
    out = tmp_file;

    info = recinfo_new();

    code = z39_present(session, "", "", "", pos, 1, info);
    if (code == NOTOK) {
        *error_message = "z39_present failed";
        retcode = -1;
    } else
        {
            retcode = 0;
#ifdef MARKER
            client_msg("-OK: %d %d\n",
                       recinfo_get_status(info),
                       recinfo_get_returned(info));
#endif
#ifndef MARKER
            client_msg("-OK: Status %d, Received %d, Position %d, (Set: %s)\n",
                       recinfo_get_status(info),
                       recinfo_get_returned(info),
                       recinfo_get_position(info),
                       session_setname(session));
#endif
        }

    recinfo_error(session, info);
    recinfo_output(session, info);
    recinfo_dispose(&info);

    fflush(out);
    out = stdout;
    (void) fclose(tmp_file);

    if (-1 == retcode)
        return(-1);

    /* now read document into docbuf */
    if (-1 == stat(WILLOW_TEMP, &stat_buf)) {
        *error_message = "couldn't stat willow temp";
        return(-1);
    }
    if ((char *)0 == (text = (char *)malloc(stat_buf.st_size+1))) {
        *error_message = "couldn't malloc for willow temp";
        return(-1);
    }
    if ((FILE *)0 == (tmp_file = fopen(WILLOW_TEMP, "r"))) {
        *error_message = "couldn't read willow temp";
        return(-1);
    }
    if (stat_buf.st_size != fread(text, 1, stat_buf.st_size, tmp_file)) {
        *error_message = "couldn't read willow temp";
        return(-1);
    }
    (void) fclose(tmp_file);
    text[stat_buf.st_size] = '\0';

    docbuf->end = 0;

    retcode = add_buf_string(text, docbuf);
    free(text);
    return(retcode);
}

/* from willow client.c find_command () */
int
run_z3950_query(dn2tosend, coll_name, int_ret_doc, num_int_ret_doc,
                                                        error_message)
char *dn2tosend, *coll_name;
INT_RETRIEVED_DOC **int_ret_doc;
int *num_int_ret_doc;
char **error_message;
{
    RECORD_INFO *info;
    int  code, retcode;
    char new_query[MAX_BUFFER*4];
    INT_RETRIEVED_DOC *ret_docs;
    int num_docs, i;
    char *names;        /* NOTE: never freed */
    char *name;

    if (-1 == smart_dn2_convert (dn2tosend, new_query)) {
        *error_message = "bad dn2 conversion";
        return(-1);
    }

    info = recinfo_new();

    code = z39_search(session, new_query, "", "", "", info);

    if (code == NOTOK)
        {
            *error_message = "z3950 search failed";
            retcode = -1;
        }
    else
        {
#ifdef MARKER
            client_msg("-OK: %d %d %d\n",
                       recinfo_get_status(info),
                       recinfo_get_result(info),
                       recinfo_get_returned(info));
#endif
#ifndef MARKER
            client_msg("-OK: Status %d, Hits %d, Received %d (Set: %s)\n",
                       recinfo_get_status(info),
                       recinfo_get_result(info),
                       recinfo_get_returned(info),
                       session_setname(session));
#endif
        }
    num_docs = recinfo_get_result(info);        /* num hits */

    recinfo_error(session, info);
    recinfo_output(session, info);    
    recinfo_dispose(&info);
    if (-1 == retcode)
        return(-1);
    
    ret_docs = (INT_RETRIEVED_DOC *)malloc(num_docs * sizeof(INT_RETRIEVED_DOC));
    if ((INT_RETRIEVED_DOC *)0 == ret_docs) {
        *error_message = "couldn't malloc ret_docs";
        return(-1);
    }

    /* assume up to 10 digit document numbers in Z3950_N */
    names = (char *)malloc(17*num_docs);
    if ((char *)0 == names) {
        *error_message = "couldn't malloc doc names";
        return(-1);
    }

    for (i = 0; i < num_docs; i++) {
        ret_docs[i].coll_name = coll_name;
        ret_docs[i].coll_type = CLTYPE_Z3950;
        name = &names[17*i];
        sprintf(name, "%.10d_Z3950", i+1);
        name[16] = '\0';
        ret_docs[i].external_name = ret_docs[i].internal_name =
                        ret_docs[i].title = name;
        ret_docs[i].score = 1.0;
        ret_docs[i].hilight_info = (char *)0;

    }

    *int_ret_doc = ret_docs;
    *num_int_ret_doc = num_docs;
}
