/*                         ATTRINFO.C                              */

/*
 *  Copyright (c) 1992 Stanford Leland Junior University
 *
 *  Commercial use prohibited
 *
 *  Author: Harold Finkbeiner (Harold@Forsythe.Stanford.Edu)
 */

/*
 *  This module allows you handle attributes information.   The attributes
 *  are expressed as a module or a value.
 */

#include <stdlib.h>

#include "attrinfo.h"
#include "bib1attr.h"
#include "z39constant.h"
#include "msg.h"

/*
 * TABLES of ATTRIBUTES
 */

ATTRIBUTE_INFO *search_info(ATTRIBUTE_TYPE *type,
			    char *text,
			    int  length)
{
    ATTRIBUTE_INFO *info = type->table;
    unsigned i;

    for (i = 0; i < type->table_size; i++, info++)
	if (memcmp(info->name, text, length) == 0)
	    return info;
    
    return NULL;
}

ATTRIBUTE_TYPE *search_type(char *text,
			    int  length)
{
    ATTRIBUTE_TYPE *type = Bib1_Attribute_Types;
    unsigned i;

    for (i = 0; i < DIM(Bib1_Attribute_Types); i++, type++)
	if (memcmp(type->name, text, length) == 0)
	    return type;

    return NULL;
}

int attribute_info(int attributes[],
		   char *attribute_start,
		   char *attribute_seperator,
		   char *attribute_end)
{
    ATTRIBUTE_TYPE *type;
    ATTRIBUTE_INFO *info;
    int result = 0;

    type = search_type(attribute_start,
		       attribute_seperator - attribute_start);

    if (type == NULL)
	result =  msg_make(BAD_ATTRIB_INFO,"unable to find a matching attribute type");
    else
	{
	    info = search_info(type,
			       attribute_seperator + 1, 
			       attribute_end - attribute_seperator - 1);

	    if (info == NULL)
		{
		    int numeric_value;
		    
		    *attribute_end = EOS;
		    
		    numeric_value = (int) strtol(attribute_seperator + 1, 
						 &attribute_end, 
						 10);
		    
		    if (*attribute_end == EOS)
			attributes[type->index] = numeric_value;
		    else
			result = msg_make(BAD_ATTRIB_INFO,"unable to find matching attribute value");
		}
	    else
		attributes[type->index] = info->value;
	}
    return result;
}


int attribute_info_parse(int attributes[],
			 char *text,
			 int  length,
			 char assignment,
			 char seperator)
{
    char ending_character;
    char *ending_location = text + length;
    char *cp;
    char *attribute_type = NULL;
    char *attribute_value = NULL;
    int  result = 0;
    
    /*
     * For efficiency instead of copying the text and appending a EOS, I just save the
     * last character and replace it when the routine is finished. 
     */
    
    ending_character = *ending_location;
    
    for (cp = text; length >= 0; length--, cp++)
	{
	   if (*cp == assignment)
	       {
		   if (attribute_type || length <= 1)
		       return msg_make(BAD_ATTRIB_INFO,"attribute value assignment incorrect");

		   if (cp != text)
		       attribute_type = text;
		   
		   attribute_value = cp;
	       }
	   else if (*cp == seperator || length == 0)
	       {
		   if (cp > text)
		       {
			   if (attribute_type == NULL)
			       return msg_make(BAD_ATTRIB_INFO,"missing the attribute type");
		   
			   result = attribute_info(attributes,
						   attribute_type,
						   attribute_value,
						   cp);
			   if (result)
			       break;

			   attribute_type = NULL;
		       }
		   text = cp + 1;
	       }
       }

    *ending_location = ending_character;
    
    return result;
}
